from typing import Type, Dict, Set, List, Tuple, Union, Any, Collection
import itertools
from multiprocessing import Pool, cpu_count
from tqdm.auto import tqdm
from loguru import logger

# This aliasing is temporary, until the schema is finalized
key_image_pointer: Type = Union[int, str, Any]
output_pointer: Type = Union[int, str, Any]
ring_vector: Type = Set[output_pointer]
ring_bucket: Type = Dict[key_image_pointer, ring_vector]
edge: Type = Dict[str, Any]


def process_bucket_single_thread_core(
    rings: ring_bucket,
    index_pairs: Union[List[Tuple[int, int]], None],
    diagnostic_level: int = 1,
    show_progress_bar: bool = True,
    verbosity_level: int = 1,
) -> List[edge]:
    """
    Core function - you do not need to interact with this directly, use `process_bucket()` below
    This takes a batch of index pairs and processes the rings in a single thread.

    :param rings: bucket of rings to analyze
    :param index_pairs: list of index pairs to process. If None provided, checks all possible combinations
    :param diagnostic_level: 0 = no diagnostics, 1 = include match_key_image_pointer
    :param show_progress_bar: show a progress bar
    :param verbosity_level: 0 = silent
    :return: identified transaction tree edges in the form of {"key_image_pointer": key_image, "output_pointer": output}
    """
    # If no index pairs are provided (could be None, (), {}, [], etc...), use all possible combinations
    if not index_pairs:
        index_pairs = list(itertools.product(list(rings.keys()), list(rings.keys())))
    if verbosity_level:
        logger.info(f"Processing {len(index_pairs)} index pairs")

    # Avoid redundant checks by only crunching the upper triangle of the index pair matrix
    key_image_pointer_pairs: Set[Tuple[key_image_pointer, key_image_pointer]] = {
        (ind1, ind2) for ind1, ind2 in tqdm(index_pairs) if ind2 > ind1
    }
    if verbosity_level:
        logger.info(f"Reduced to {len(key_image_pointer_pairs)} index pairs")

    # Process the bucket
    edges: List[edge] = []
    if show_progress_bar:
        p = tqdm(key_image_pointer_pairs, mininterval=1)
    else:
        p = key_image_pointer_pairs
    for key_image_pointer_left, key_image_pointer_right in p:
        ring_left: ring_vector = rings[key_image_pointer_left]
        ring_right: ring_vector = rings[key_image_pointer_right]
        # Do we have a singleton?
        if len(ring_left & ring_right) == len(ring_left) - 1:
            left_edge: Dict[str, Any] = {
                "key_image_pointer": key_image_pointer_left,
                "output_pointer": (ring_left - ring_right).pop(),
            }
            if diagnostic_level > 0:
                left_edge["match_key_image_pointer"] = key_image_pointer_right
            edges.append(left_edge)

            right_edge: Dict[str, Any] = {
                "key_image_pointer": key_image_pointer_right,
                "output_pointer": (ring_right - ring_left).pop(),
            }
            if diagnostic_level > 0:
                right_edge["match_key_image_pointer"] = key_image_pointer_left
            edges.append(right_edge)

    return edges


def process_bucket(
    rings: ring_bucket,
    index_pairs: Union[None, Collection[Tuple[int, int]]] = None,
    num_workers: Union[int, None] = None,
    diagnostic_level: int = 1,
    verbosity_level: int = 1,
) -> List[edge]:
    """
    This function takes a bucket of rings and processes them to identify transaction tree edges from singletons
    that arise from XORing ring pairs (i.e. symmetric set difference)

    :param rings: bucket of rings to analyze
    :param index_pairs: list of index pairs to process. If None provided, checks all possible combinations
    :param num_workers: number of workers to use. If None provided, uses all available cores
    :param diagnostic_level: 0 = no diagnostics, 1 = include match_key_image_pointer
    :param verbosity_level: 0 = silent
    :return: identified transaction tree edges in the form of {"key_image_pointer": key_image, "output_pointer": output}
    """
    if verbosity_level:
        logger.info(f"Processing {len(rings)} rings")

    # If no index pairs are provided, use all possible combinations
    if index_pairs is None:
        index_pairs = list(itertools.product(list(rings.keys()), list(rings.keys())))
    if verbosity_level:
        logger.info(f"Processing {len(index_pairs)} index pairs")

    # Avoid redundant checks by only crunching the upper triangle of the index pair matrix
    key_image_pointer_pairs: List[Tuple[key_image_pointer, key_image_pointer]] = [
        (ind1, ind2) for ind1, ind2 in tqdm(index_pairs, mininterval=1) if ind2 > ind1
    ]
    if verbosity_level:
        logger.info(f"Reduced to {len(key_image_pointer_pairs)} index pairs")

    # Use all available cores unless specified otherwise
    if num_workers is None:
        num_workers: int = cpu_count()

    if num_workers <= 1:
        # If we only have one worker, just run the single-threaded version
        return process_bucket_single_thread_core(rings, index_pairs=key_image_pointer_pairs)
    else:
        # Split the work into chunks
        batches = [list(key_image_pointer_pairs)[i::num_workers] for i in range(num_workers)]
        if verbosity_level:
            logger.info(f"Split work into {num_workers} batches")
        iterable = [(rings, batch, diagnostic_level) for batch in batches]
        if verbosity_level:
            logger.info(f"Starting parallel processing with {num_workers} workers")

        # Process the chunks in parallel
        with Pool(processes=num_workers) as pool:
            results: List[List[edge]] = pool.starmap(
                process_bucket_single_thread_core,
                iterable,
            )
        if verbosity_level:
            logger.info(f"Processed {len(results)} result batches")

        # Combine the results
        edges: List[edge] = []
        for result in results:
            edges.extend(result)
        if verbosity_level:
            logger.info(f"Reshaped to {len(edges)} edges")

        return edges



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext
import re

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")) as fh:
        return fh.read()


setup(
    name="ringxor",
    version="0.0.3.1a",
    license="MIT",
    description="Tooling for passive deanonymization of Monero-style blockchains using differ-by-one ring comparisons",
    long_description=read("README.rst"),
    author="Isthmus (Mitchell P. Krawiec-Thayer)",
    author_email="ringxor@mitchellpkt.com",
    url="https://github.com/mitchellpkt/ringxor",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Utilities",
    ],
    project_urls={
        "Changelog": "https://github.com/mitchellpkt/ringxor/blob/main/CHANGELOG.rst",
        "Issue Tracker": "https://github.com/mitchellpkt/ringxor/issues",
    },
    keywords=[
        "monero",
        "chainanalysis",
        "chainalysis",
        "blockchain",
        "deanonymization",
        "ringxor",
    ],
    python_requires=">=3.8",
    install_requires=[
        "tqdm",
        "loguru",
    ],
    extras_require={},
    # entry_points={
    #     "console_scripts": [
    #         "isthmuslib = isthmuslib.cli:main",
    #     ]
    # },
)



import json
from multiprocessing import cpu_count
from src.ringxor import ringxor
import pathlib

# Load the test data
data_path: pathlib.Path = pathlib.Path.cwd() / ".." / "data" / "version_controlled" / "demo_rings.json"
with open(data_path, "r") as f:
    all_rings_raw = json.load(f)

# The code uses sets, but they can't be serialized to json, so we convert back here from temporary list representation
all_rings: ringxor.ring_bucket = {key_image: set(ring) for key_image, ring in all_rings_raw.items()}


def test_ringxor_process_bucket_single_thread_core():
    results = sorted(
        ringxor.process_bucket_single_thread_core(all_rings, index_pairs=None),
        key=lambda x: x["key_image_pointer"],
    )

    assert len(results) == 54
    assert results[0] == {
        "key_image_pointer": "056189f0c3ed7806bc17655152613b3288dfa496b6566b9bbef930efaa87975f",
        "output_pointer": "416e9438f9f69410cf42959dc8ee515318ea6710e5bf8eb76f69c6183a09bded",
        "match_key_image_pointer": "af6cb1936c08108500c3756e81073b6e48b33405be54df891e8b6e433b66b216",
    }
    assert results[53] == {
        "key_image_pointer": "fdfcbe3f85e480905d5db681bf89e5b81dc514871f0735c8a8cc7dffb6d7cde8",
        "output_pointer": "6d1fe6224bfccbb5f660d08bdb23fba89e75cf7d367b16bc40bcb3d337e12b0e",
        "match_key_image_pointer": "7d64ac02c2823257fb88ab2b34c2ecb6fbd60c59a23650d682d42c71105ac32a",
    }


def test_ringxor_process_bucket_1_worker():
    results = sorted(
        ringxor.process_bucket(all_rings, index_pairs=None, num_workers=1),
        key=lambda x: x["key_image_pointer"],
    )

    assert len(results) == 54
    assert results[0] == {
        "key_image_pointer": "056189f0c3ed7806bc17655152613b3288dfa496b6566b9bbef930efaa87975f",
        "output_pointer": "416e9438f9f69410cf42959dc8ee515318ea6710e5bf8eb76f69c6183a09bded",
        "match_key_image_pointer": "af6cb1936c08108500c3756e81073b6e48b33405be54df891e8b6e433b66b216",
    }
    assert results[53] == {
        "key_image_pointer": "fdfcbe3f85e480905d5db681bf89e5b81dc514871f0735c8a8cc7dffb6d7cde8",
        "output_pointer": "6d1fe6224bfccbb5f660d08bdb23fba89e75cf7d367b16bc40bcb3d337e12b0e",
        "match_key_image_pointer": "7d64ac02c2823257fb88ab2b34c2ecb6fbd60c59a23650d682d42c71105ac32a",
    }


def test_ringxor_process_bucket_N_worker():
    if cpu_count() == 1:
        print("Skipping test_ringxor_process_bucket_N_worker because only 1 CPU is available")
    else:
        results = sorted(
            ringxor.process_bucket(all_rings, index_pairs=None, num_workers=2),
            key=lambda x: x["key_image_pointer"],
        )

        assert len(results) == 54
        assert results[0] == {
            "key_image_pointer": "056189f0c3ed7806bc17655152613b3288dfa496b6566b9bbef930efaa87975f",
            "output_pointer": "416e9438f9f69410cf42959dc8ee515318ea6710e5bf8eb76f69c6183a09bded",
            "match_key_image_pointer": "af6cb1936c08108500c3756e81073b6e48b33405be54df891e8b6e433b66b216",
        }
        assert results[53] == {
            "key_image_pointer": "fdfcbe3f85e480905d5db681bf89e5b81dc514871f0735c8a8cc7dffb6d7cde8",
            "output_pointer": "6d1fe6224bfccbb5f660d08bdb23fba89e75cf7d367b16bc40bcb3d337e12b0e",
            "match_key_image_pointer": "7d64ac02c2823257fb88ab2b34c2ecb6fbd60c59a23650d682d42c71105ac32a",
        }


def test_ringxor_process_bucket_diagnostic_level():
    results = sorted(
        ringxor.process_bucket(all_rings, index_pairs=None, diagnostic_level=1),
        key=lambda x: x["key_image_pointer"],
    )
    assert len(results) == 54

    # Check 0th element
    assert results[0] == {
        "key_image_pointer": "056189f0c3ed7806bc17655152613b3288dfa496b6566b9bbef930efaa87975f",
        "output_pointer": "416e9438f9f69410cf42959dc8ee515318ea6710e5bf8eb76f69c6183a09bded",
        "match_key_image_pointer": "af6cb1936c08108500c3756e81073b6e48b33405be54df891e8b6e433b66b216",
    }

    # Check counterpart of 0th element
    counterpart: dict = next(
        x for x in results if x["key_image_pointer"] == results[0]["match_key_image_pointer"]
    )
    assert counterpart["match_key_image_pointer"] == results[0]["key_image_pointer"]
